@@ -29,6 +29,7 @@
 #include <config.h>
 #endif
 
+#include <limits.h>
 #include <stdio.h>
 #include <X11/Xlib.h>
 /* we need to be able to manipulate the Display structure on events */
@@ -82,7 +83,7 @@ Time XRRConfigTimes (XRRScreenConfiguration *config, Time *config_timestamp)
 }
 
 
-SizeID XRRConfigCurrentConfiguration (XRRScreenConfiguration *config, 
+SizeID XRRConfigCurrentConfiguration (XRRScreenConfiguration *config,
 			      Rotation *rotation)
 {
     *rotation = (Rotation) config->current_rotation;
@@ -94,20 +95,17 @@ short XRRConfigCurrentRate (XRRScreenConfiguration *config)
     return config->current_rate;
 }
 
-/* 
- * Go get the screen configuration data and salt it away for future use; 
+/*
+ * Go get the screen configuration data and salt it away for future use;
  * returns NULL if extension not supported
  */
-static XRRScreenConfiguration *_XRRValidateCache (Display *dpy, 
+static XRRScreenConfiguration *_XRRValidateCache (Display *dpy,
 						  XExtDisplayInfo *info,
 						  int screen)
 {
-    XRRScreenConfiguration **configs;
-    XRandRInfo *xrri;
-
     if ((screen >= 0) && (screen < ScreenCount(dpy)) && XextHasExtension(info)) {
-	xrri = (XRandRInfo *) info->data;
-	configs = xrri->config;
+	XRandRInfo *xrri = (XRandRInfo *) info->data;
+	XRRScreenConfiguration **configs = xrri->config;
 
 	if (configs[screen] == NULL)
 	    configs[screen] = _XRRGetScreenInfo (dpy, info, RootWindow(dpy, screen));
@@ -122,11 +120,11 @@ Rotation XRRRotations(Display *dpy, int screen, Rotation *current_rotation)
 {
   XRRScreenConfiguration *config;
   XExtDisplayInfo *info = XRRFindDisplay(dpy);
-  Rotation cr;
+
   LockDisplay(dpy);
   if ((config = _XRRValidateCache(dpy, info, screen))) {
+    Rotation cr = config->rotations;
     *current_rotation = config->current_rotation;
-    cr = config->rotations;
     UnlockDisplay(dpy);
     return cr;
   }
@@ -140,14 +138,13 @@ Rotation XRRRotations(Display *dpy, int screen, Rotation *current_rotation)
 /* given a screen, return the information from the (possibly) cached data */
 XRRScreenSize *XRRSizes(Display *dpy, int screen, int *nsizes)
 {
-  XRRScreenConfiguration *config; 
+  XRRScreenConfiguration *config;
   XExtDisplayInfo *info = XRRFindDisplay(dpy);
-  XRRScreenSize *sizes;
 
   LockDisplay(dpy);
   if ((config = _XRRValidateCache(dpy, info, screen))) {
+    XRRScreenSize *sizes = config->sizes;
     *nsizes = config->nsizes;
-    sizes = config->sizes;
     UnlockDisplay(dpy);
     return sizes;
     }
@@ -155,18 +152,17 @@ XRRScreenSize *XRRSizes(Display *dpy, int screen, int *nsizes)
     UnlockDisplay(dpy);
     *nsizes = 0;
     return NULL;
-  }  
+  }
 }
 
 short *XRRRates (Display *dpy, int screen, int sizeID, int *nrates)
 {
-  XRRScreenConfiguration *config; 
+  XRRScreenConfiguration *config;
   XExtDisplayInfo *info = XRRFindDisplay(dpy);
-  short *rates;
 
   LockDisplay(dpy);
   if ((config = _XRRValidateCache(dpy, info, screen))) {
-    rates = XRRConfigRates (config, sizeID, nrates);
+    short *rates = XRRConfigRates (config, sizeID, nrates);
     UnlockDisplay(dpy);
     return rates;
     }
@@ -174,20 +170,19 @@ short *XRRRates (Display *dpy, int screen, int sizeID, int *nrates)
     UnlockDisplay(dpy);
     *nrates = 0;
     return NULL;
-  }  
+  }
 }
 
 /* given a screen, return the information from the (possibly) cached data */
 Time XRRTimes (Display *dpy, int screen, Time *config_timestamp)
 {
-  XRRScreenConfiguration *config; 
+  XRRScreenConfiguration *config;
   XExtDisplayInfo *info = XRRFindDisplay(dpy);
-  Time ts;
 
   LockDisplay(dpy);
   if ((config = _XRRValidateCache(dpy, info, screen))) {
+      Time ts = config->timestamp;
       *config_timestamp = config->config_timestamp;
-      ts = config->timestamp;
       UnlockDisplay(dpy);
       return ts;
     } else {
@@ -206,12 +201,10 @@ static XRRScreenConfiguration *_XRRGetScreenInfo (Display *dpy,
     _XAsyncHandler 	    async;
     _XRRVersionState	    async_state;
     int			    nbytes, nbytesRead, rbytes;
-    int			    i;
     xScreenSizes	    size;
     struct _XRRScreenConfiguration  *scp;
     XRRScreenSize	    *ssp;
     short    		    *rates;
-    xRRQueryVersionReq      *vreq;
     XRandRInfo		    *xrri;
     Bool		    getting_version = False;
 
@@ -221,13 +214,15 @@ static XRRScreenConfiguration *_XRRGetScreenInfo (Display *dpy,
 
     if (xrri->major_version == -1)
     {
+	xRRQueryVersionReq	*vreq;
+
 	/* hide a version query in the request */
 	GetReq (RRQueryVersion, vreq);
 	vreq->reqType = info->codes->major_opcode;
 	vreq->randrReqType = X_RRQueryVersion;
 	vreq->majorVersion = RANDR_MAJOR;
 	vreq->minorVersion = RANDR_MINOR;
-    
+
 	async_state.version_seq = dpy->request;
 	async_state.error = False;
 	async.next = dpy->async_handlers;
@@ -272,23 +267,30 @@ static XRRScreenConfiguration *_XRRGetScreenInfo (Display *dpy,
 	rep.rate = 0;
 	rep.nrateEnts = 0;
     }
-    
-    nbytes = (long) rep.length << 2;
+    if (rep.length < INT_MAX >> 2) {
+	nbytes = (long) rep.length << 2;
 
-    nbytesRead = (long) (rep.nSizes * SIZEOF (xScreenSizes) +
-			 ((rep.nrateEnts + 1)& ~1) * 2 /* SIZEOF (CARD16) */);
-    
-    /* 
-     * first we must compute how much space to allocate for 
-     * randr library's use; we'll allocate the structures in a single
-     * allocation, on cleanlyness grounds.
-     */
+	nbytesRead = (long) (rep.nSizes * SIZEOF (xScreenSizes) +
+			    ((rep.nrateEnts + 1)& ~1) * 2 /* SIZEOF(CARD16) */);
 
-    rbytes = sizeof (XRRScreenConfiguration) +
-      (rep.nSizes * sizeof (XRRScreenSize) +
-       rep.nrateEnts * sizeof (int));
+	/*
+	 * first we must compute how much space to allocate for
+	 * randr library's use; we'll allocate the structures in a single
+	 * allocation, on cleanlyness grounds.
+	 */
+
+	rbytes = sizeof (XRRScreenConfiguration) +
+	  (rep.nSizes * sizeof (XRRScreenSize) +
+	   rep.nrateEnts * sizeof (int));
+
+	scp = Xmalloc(rbytes);
+    } else {
+	nbytes = 0;
+	nbytesRead = 0;
+	rbytes = 0;
+	scp = NULL;
+    }
 
-    scp = (struct _XRRScreenConfiguration *) Xmalloc(rbytes);
     if (scp == NULL) {
 	_XEatData (dpy, (unsigned long) nbytes);
 	return NULL;
@@ -299,7 +301,7 @@ static XRRScreenConfiguration *_XRRGetScreenInfo (Display *dpy,
     rates = (short *) (ssp + rep.nSizes);
 
     /* set up the screen configuration structure */
-    scp->screen = 
+    scp->screen =
       ScreenOfDisplay (dpy, XRRRootToScreen(dpy, rep.root));
 
     scp->sizes = ssp;
@@ -320,9 +322,9 @@ static XRRScreenConfiguration *_XRRGetScreenInfo (Display *dpy,
     /*
      * First the size information
      */
-    for (i = 0; i < rep.nSizes; i++)  {
+    for (unsigned int i = 0; i < rep.nSizes; i++)  {
 	_XReadPad (dpy, (char *) &size, SIZEOF (xScreenSizes));
-	
+
         ssp[i].width = size.widthInPixels;
 	ssp[i].height = size.heightInPixels;
 	ssp[i].mwidth = size.widthInMillimeters;
@@ -332,13 +334,13 @@ static XRRScreenConfiguration *_XRRGetScreenInfo (Display *dpy,
      * And the rates
      */
     _XRead16Pad (dpy, rates, 2 /* SIZEOF (CARD16) */ * rep.nrateEnts);
-    
+
     /*
      * Skip any extra data
      */
     if (nbytes > nbytesRead)
 	_XEatData (dpy, (unsigned long) (nbytes - nbytesRead));
-    
+
     return (XRRScreenConfiguration *)(scp);
 }
 
@@ -354,7 +356,7 @@ XRRScreenConfiguration *XRRGetScreenInfo (Display *dpy, Window window)
     return config;
 }
 
-    
+
 void XRRFreeScreenConfigInfo (XRRScreenConfiguration *config)
 {
     Xfree (config);
@@ -364,7 +366,7 @@ Status XRRSetScreenConfigAndRate (Display *dpy,
 				  XRRScreenConfiguration *config,
 				  Drawable draw,
 				  int size_index,
-				  Rotation rotation, 
+				  Rotation rotation,
 				  short rate,
 				  Time timestamp)
 {
@@ -378,7 +380,7 @@ Status XRRSetScreenConfigAndRate (Display *dpy,
     /* Make sure has_rates is set */
     if (!XRRQueryVersion (dpy, &major, &minor))
 	return 0;
-    
+
     LockDisplay (dpy);
     xrri = (XRandRInfo *) info->data;
     if (xrri->has_rates)
@@ -406,7 +408,7 @@ Status XRRSetScreenConfigAndRate (Display *dpy,
 	req->timestamp = timestamp;
 	req->configTimestamp = config->config_timestamp;
     }
-    
+
     (void) _XReply (dpy, (xReply *) &rep, 0, xTrue);
 
     /* actually .errorCode in struct xError */
